@@ -46,7 +46,7 @@ struct mail_user {
 	const char *service;
 	const char *session_id;
 	struct mail_user_connection_data conn;
-	const char *auth_token, *auth_user;
+	const char *auth_mech, *auth_token, *auth_user;
 	const char *const *userdb_fields;
 	/* Timestamp when this session was initially created. Most importantly
 	   this stays the same after IMAP client is hibernated and restored. */
@@ -136,6 +136,8 @@ int mail_user_init(struct mail_user *user, const char **error_r);
 
 void mail_user_ref(struct mail_user *user);
 void mail_user_unref(struct mail_user **user);
+/* Assert that this is the last reference for the user and unref it. */
+void mail_user_deinit(struct mail_user **user);
 
 /* Duplicate a mail_user. mail_user_init() and mail_namespaces_init() need to
    be called before the user is usable. */
@@ -215,4 +217,28 @@ void mail_user_stats_fill(struct mail_user *user, struct stats *stats);
    with ENOENT. This way it avoids unnecessary disk IO to the home. */
 int mail_user_home_mkdir(struct mail_user *user);
 
+/* Obtain the postmaster address to be used for this user as an RFC 5322 (IMF)
+   address. Returns false if the configured postmaster address is invalid in
+   which case error_r contains the error message. */
+static inline bool
+mail_user_get_postmaster_address(struct mail_user *user,
+				 const struct message_address **address_r,
+				 const char **error_r)
+{
+	return mail_user_set_get_postmaster_address(user->set, address_r,
+						    error_r);
+}
+
+/* Obtain the postmaster address to be used for this user as an RFC 5321 (SMTP)
+   address. Returns false if the configured postmaster address is invalid in
+   which case error_r contains the error message. */
+static inline bool
+mail_user_get_postmaster_smtp(struct mail_user *user,
+			      const struct smtp_address **address_r,
+			      const char **error_r)
+{
+	return mail_user_set_get_postmaster_smtp(user->set, address_r,
+						 error_r);
+}
+
 #endif
